/*-
 * Copyright (c) 1998 Alex Nash
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: wdt.c,v 1.1.1.1 1998/07/03 23:14:42 alex Exp $
 */

#include "wdt.h"

#if NWDT > 0

#ifndef ACTUALLY_LKM_NOT_KERNEL
#include "opt_wdt.h"
#endif

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/callout.h>
#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/reboot.h>
#include <sys/proc.h>
#include <sys/lkm.h>

#include <i386/isa/isa.h>
#include <i386/isa/isa_device.h>
#include <i386/isa/isa_device.h>
#include <i386/isa/wdt.h>

#define CDEV_MAJOR			14		/* major device number				*/

/*
 *	8254 counter mappings
 */
#define WDT_8254_TC_LO		0		/* low 16 bits of timeout counter	*/
#define	WDT_8254_TC_HI		1		/* high 16 bits of timeout counter	*/
#define WDT_8254_BUZZER		2

/*
 *	WDT500/501 ports
 */
#define WDT_8254_BASE		(wdp->iobase)
#define WDT_8254_CTL		(WDT_8254_BASE + 3)
#define WDT_DISABLE_TIMER	(wdp->iobase + 7)
#define WDT_ENABLE_TIMER	(wdp->iobase + 7)

/*
 *	WDT501 specific ports
 */
#define WDT_STATUS_REG		(wdp->iobase + 4)
#define WDT_START_BUZZER	(wdp->iobase + 4)
#define WDT_TEMPERATURE		(wdp->iobase + 5)
#define WDT_STOP_BUZZER		(wdp->iobase + 5)

/*
 *	watchdog
 *
 *	State information for each wdt device in the system.
 */
struct watchdog
{
	int			unit;			/* wd unit number				*/
	int			iobase;			/* I/O base address				*/
	unsigned	timeout_secs;	/* watchdog timeout in seconds	*/
	unsigned	procs;			/* processes sleeping on wdschd */
	int			busy;			/* true if device is open		*/
};

static struct watchdog wd[NWDT];

/*
 *	function prototypes for wdt_cdevsw
 */
static int wdtopen (dev_t dev, int flags, int fmt, struct proc *p);
static int wdtclose (dev_t dev, int flags, int fmt, struct proc *p);
static int wdtioctl (dev_t dev, int cmd, caddr_t data, int flag,
					 struct proc *p);

/*
 *	function prototypes for wdtdriver
 */
static int wdtprobe (struct isa_device *dev);
static int wdtattach (struct isa_device *dev);

/*
 *	wdt_cdevsw
 */
static struct cdevsw wdt_cdevsw = 
{
	wdtopen,	wdtclose,	noread,		nowrite,
	wdtioctl,	nostop,		nullreset,	nodevtotty,
	noselect,	nommap,		NULL,		"wdt",
	NULL,		-1
};

/*
 *	wdtdriver
 */
struct isa_driver wdtdriver =
{
	wdtprobe,
	wdtattach,
	"wdt",
	1
};

/*
 *	wdt_8254_count
 *
 *	Loads the specified counter with the 16-bit value 'v'.
 */
static void
wdt_8254_count (struct watchdog *wdp, int counter, u_int16_t v)
{
	outb(WDT_8254_BASE + counter, v & 0xFF);
	outb(WDT_8254_BASE + counter, v >> 8);
}

/*
 *	wdt_8254_mode
 *
 *	Sets the mode of the specified counter.
 */
static void
wdt_8254_mode (struct watchdog *wdp, int counter, int mode)
{
	outb(WDT_8254_CTL, (counter << 6) | 0x30 | (mode << 1));
}

/*
 *	wdt_set_timeout
 *
 *	Load the watchdog timer with the specified number of seconds.
 */
static void
wdt_set_timeout (struct watchdog *wdp, unsigned seconds)
{
	u_int16_t v = (u_int16_t)seconds * 14;

	inb(WDT_DISABLE_TIMER);
	wdt_8254_count(wdp, WDT_8254_TC_HI, v);
	outb(WDT_ENABLE_TIMER, 0);
}

/*
 *	wdt_timeout
 *
 *	Kernel timeout handler.  This function is called every
 *	wdp->timeout_secs / 2 seconds.  It reloads the watchdog
 *	counters in one of two ways:
 *
 *	   - If there are one or more processes sleeping in a
 *	     WIOCSCHED ioctl(), they are woken up to perform
 *	     the counter reload.
 *	   - If no processes are sleeping in WIOCSCHED, the
 *	     counters are reloaded from here.
 *
 *	Finally, another timeout is scheduled for wdp->timeout_secs
 *	from now.
 */
static void
wdt_timeout (void *arg)
{
	struct watchdog *wdp = (struct watchdog *)arg;

	/* reload counters from proc in WIOCSCHED ioctl()? */
	if ( wdp->procs )
		wakeup_one(wdp);
	else
		wdt_set_timeout(wdp, wdp->timeout_secs);

	/* schedule another timeout in half the countdown time */
	timeout(wdt_timeout, arg, wdp->timeout_secs * hz / 2);
}

/*
 *	wdt_timer_disable
 *
 *	Disables the watchdog timer and cancels the scheduled (if any)
 *	kernel timeout.
 */
static void
wdt_timer_disable (struct watchdog *wdp)
{
	inb(WDT_DISABLE_TIMER);
	untimeout(wdt_timeout, wdp);
}

/*
 *	wdt_timer_enable
 *
 *	Enables the watchdog timer to expire in the specified number
 *	of seconds.  If 'seconds' is outside the range 2-4680, it
 *	is silently clamped to be within range.
 */
static void
wdt_timer_enable (struct watchdog *wdp, unsigned seconds)
{
	int s;

	/* clamp range */
	if ( seconds < 2 )
		seconds = 2;

	if ( seconds > 4680 )
		seconds = 4680;

	/* block out the timeout handler */
	s = splclock();

	wdt_timer_disable(wdp);
	wdp->timeout_secs = seconds;
	timeout(wdt_timeout, wdp, hz * seconds / 2);
	wdt_set_timeout(wdp, seconds);

	/* re-enable clock interrupts */
	splx(s);
}

/*
 *	wdt_init_timer
 *
 *	Configure the modes for the watchdog counters and initialize
 *	the low 16-bits of the watchdog counter to have a period of
 *	approximately 1/14th of a second.
 */
static void
wdt_init_timer (struct watchdog *wdp)
{
	wdt_8254_mode(wdp, WDT_8254_TC_LO, 3);
	wdt_8254_mode(wdp, WDT_8254_TC_HI, 2);
	wdt_8254_count(wdp, WDT_8254_TC_LO, 63920);
}

/*******************************************************************
 *	WDT501 specific functions
 *******************************************************************/

#ifdef WDT501

/*
 *	wdt_buzzer_off
 *
 *	Turns the buzzer off.
 */
static void
wdt_buzzer_off (struct watchdog *wdp)
{
	outb(WDT_STOP_BUZZER, 0);
}

/*
 *	wdt_buzzer_enable
 *
 *	Enables the buzzer when the watchdog counter expires.
 */
static void
wdt_buzzer_enable (struct watchdog *wdp)
{
	wdt_8254_mode(wdp, WDT_8254_BUZZER, 1);
}

/*
 *	wdt_buzzer_disable
 *
 *	Disables the buzzer from sounding when the watchdog counter
 *	expires.
 */
static void
wdt_buzzer_disable (struct watchdog *wdp)
{
	wdt_8254_mode(wdp, WDT_8254_BUZZER, 0);
}

/*
 *	wdt_read_temperature
 *
 *	Returns the temperature (in Fahrenheit) from the board.
 */
static int
wdt_read_temperature (struct watchdog *wdp)
{
	unsigned v = inb(WDT_TEMPERATURE);

	return( (v * 11) / 15 + 7 );
}

/*
 *	wdt_read_status
 *
 *	Returns the status register bits minus the counter refresh
 *	and IRQ generated bits.
 */
static int
wdt_read_status (struct watchdog *wdp)
{
	/* mask off counter refresh & IRQ generated bits */
	return(inb(WDT_STATUS_REG) & 0x7E);
}

/*
 *	wdt_display_status
 *
 *	Displays the current timeout, temperature, and power supply
 *	over/undervoltages to the console.
 */
static void
wdt_display_status (struct watchdog *wdp)
{
	int status = wdt_read_status(wdp);
	int temp   = wdt_read_temperature(wdp);

	printf("wdt%d: timeout %d secs, temp %d F",
		   wdp->unit, wdp->timeout_secs, temp);

	/* overvoltage bit is active low */
	if ( (status & WDT_SR_PS_OVER) == 0 )
		printf(" <PS overvoltage>");

	/* undervoltage bit is active low */
	if ( (status & WDT_SR_PS_UNDER) == 0 )
		printf(" <PS undervoltage>");

	printf("\n");
}

/*
 *	wdt_get_state
 *
 *	Returns the temperature and status bits.
 */
static int
wdt_get_state (struct watchdog *wdp, struct wdt_state *state)
{
	state->temperature	= wdt_read_temperature(wdp);
	state->status		= wdt_read_status(wdp);

	return(0);
}

#endif /* defined(WDT501) */

/*
 *	wdt_shutdown
 *
 *	Disables the watchdog timer at system shutdown time.
 */
static void
wdt_shutdown (int howto, void *arg)
{
	struct watchdog *wdp = (struct watchdog *)arg;

	wdt_timer_disable(wdp);
}

/*
 *	wdt_sched
 *
 *	Put the process into an infinite loop in which:
 *
 *	  - The process sleeps, waiting for a wakeup() from the timeout()
 *	    handler.
 *	  - When awakened, the process reloads the watchdog counter and
 *	    repeats the loop.
 *
 *	The only way the loop can be broken is if the process is interrupted
 *	via a signal.
 *
 *	The whole point of this is to cause a watchdog timeout to be
 *	generated if processes are no longer being scheduled.
 */
static int
wdt_sched (struct watchdog *wdp, struct proc *p)
{
	int error;
	int s;

	/*
	 * Regardless of the device permissions, you must be
	 * root to do this -- a process which is STOPPED
	 * while in this function can cause a reboot to occur
	 * if the counters aren't reloaded within wdp->timeout_secs
	 * seconds.
	 */
	if ( ( error = suser(p->p_ucred, &p->p_acflag) ) )
		return(error);

	/* block out the timeout handler */
	s = splclock();

	/* indicate that we are sleeping */
	++wdp->procs;

	while ( 1 )
	{
		error = tsleep(wdp, PCATCH | PSWP, "wdtsch", 0);

		wdt_set_timeout(wdp, wdp->timeout_secs);

		if ( error != 0 )
			break;
	}

	/* remove sleeping indication */
	--wdp->procs;

	/* re-enable timeout handler */
	splx(s);

	return(error);
}

/*******************************************************************
 *	Device ops
 *******************************************************************/

/*
 *	wdtopen
 *
 *	Device open() entry point.
 */
static int
wdtopen (dev_t dev, int flags, int fmt, struct proc *p)
{
	int minor = minor(dev);

	if ( minor >= NWDT )
		return(ENXIO);

	wd[minor].busy = 1;

	return(0);
}

/*
 *	wdtclose
 *
 *	Device close() entry point.
 */
static int
wdtclose (dev_t dev, int flags, int fmt, struct proc *p)
{
	wd[minor(dev)].busy = 0;

	return(0);
}

/*
 *	wdtioctl
 *
 *	Device ioctl() entry point.
 */
static int
wdtioctl (dev_t dev, int cmd, caddr_t data, int flag, struct proc *p)
{
	struct watchdog *	wdp		= &wd[minor(dev)];
	int					error 	= 0;

	switch ( cmd )
	{
		case WIOCSCHED:
			error = wdt_sched(wdp, p);
			break;

#ifdef WDT501
		case WIOCGETSTATE:
			error = wdt_get_state(wdp, (struct wdt_state *)data);
			break;
#endif

		default:
			error = ENXIO;
			break;
	}

	return(error);
}

/*******************************************************************
 *	ISA probe & attach functions
 *******************************************************************/

/*
 *	wdtprobe
 *
 *	ISA device probe routine.
 */
static int
wdtprobe (struct isa_device *dev)
{
	/* XXX: should talk to the 8254 to verify its presence */
	return(1);
}

/*
 *	wdattach
 *
 *	ISA device attach routine.
 */
static int
wdtattach (struct isa_device *dev)
{
	struct watchdog *wdp = &wd[dev->id_unit];

	/*
	 * register a callback for system shutdown
	 * (we need to disable the watchdog timer then)
	 */
	if ( at_shutdown(wdt_shutdown, wdp, SHUTDOWN_POST_SYNC) != 0 )
		return(0);

	/* initialize the 'watchdog' structure */
	wdp->procs	= 0;
	wdp->unit	= dev->id_unit;
	wdp->iobase = dev->id_iobase;

#ifdef WDT501
	/* turn off the buzzer, it may have been activated by a previous timeout */
	wdt_buzzer_off(wdp);

# ifdef WDT_ENABLE_BUZZER
	wdt_buzzer_enable(wdp);
# else
	wdt_buzzer_disable(wdp);
# endif
#endif

	/* initialize the timer modes and the lower 16-bit counter */
	wdt_init_timer(wdp);

	wdt_timer_enable(wdp, 30/*seconds*/);

#ifdef WDT501
	/* display temperature and powersupply problems (if any) */
	wdt_display_status(wdp);
#endif

	return(1);
}

/*******************************************************************
 *	Non-LKM code
 *******************************************************************/
#ifndef ACTUALLY_LKM_NOT_KERNEL

/*
 *	wdt_drvinit
 *
 *	Hook our cdevsw into the system.
 */
static void
wdt_drvinit (void *unused)
{
	dev_t dev;

	dev = makedev(CDEV_MAJOR, 0);
	cdevsw_add(&dev, &wdt_cdevsw, NULL);
}

SYSINIT(wdtdev,
		SI_SUB_DRIVERS,
		SI_ORDER_FIRST + CDEV_MAJOR,
		wdt_drvinit,
		NULL)

#endif	/* !defined(ACTUALLY_LKM_NOT_KERNEL) */

/*******************************************************************
 *	LKM only code
 *******************************************************************/

#ifdef ACTUALLY_LKM_NOT_KERNEL

MOD_DEV(wdt, LM_DT_CHAR, CDEV_MAJOR, &wdt_cdevsw);

static struct isa_device dev =
{
	0,					/* device id					*/
	&wdtdriver,			/* ISA driver struct			*/
	WDT_IO_BASE,		/* base I/O address				*/
	-1,					/* interrupt request			*/
	-1,					/* DMA request					*/
	(caddr_t) 0,		/* physical I/O memory			*/
	0,					/* size of I/O memory			*/
	0,					/* interrupt handler routine	*/
	0,					/* unit number					*/
	0,					/* flags						*/
	0,					/* SCSI ID						*/
	0,					/* device is present			*/
	0,					/* flags for register_intr		*/
	0,					/* hot eject 					*/
	1,					/* is device enabled			*/
	0,					/* allow conflict?				*/
	0					/* next device pointer			*/
};

/*
 *	wdt_load
 *
 *	Entry point called after the LKM is loaded into memory.
 */
static int
wdt_load (struct lkm_table *lkmtp, int cmd)
{
	if ( wdtprobe(&dev) == 0 )
	{
		uprintf("wdt: probe failed\n");
		return(ENXIO);
	}

	if ( wdtattach(&dev) == 0 )
	{
		uprintf("wdt: attach failed\n");
		return(ENXIO);
	}

	return(0);
}

/*
 *	wdt_unload
 *
 *	Called when unloading the LKM.
 */
static int
wdt_unload (struct lkm_table *lkmtp, int cmd)
{
	struct watchdog *wdp = &wd[0];

	/* if someone has the device open, return EBUSY */
	if ( wdp->busy )
		return(EBUSY);

	/* turn off the watchdog timer -- we won't be around to reload it */
	wdt_timer_disable(wdp);

	/* remove our shutdown handler */
	rm_at_shutdown(wdt_shutdown, wdp);

	return(0);
}

/*
 *	wdt_mod
 *
 *	General purpose entry point into the module.
 */
int
wdt_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
	MOD_DISPATCH(wdt, lkmtp, cmd, ver,
		wdt_load, wdt_unload, lkm_nullcmd);
}

#endif /* ACTUALLY_LKM_NOT_KERNEL */

#endif /* NWDT > 0 */
